@@ -1,8 +1,9 @@
 package Class::StateMachine;
 
-our $VERSION = '0.21';
+our $VERSION = '0.22';
 
-our $debug //= 0;
+our $debug                     //= 0;
+our $ignore_same_state_changes //= 0;
 
 package Class::StateMachine::Private;
 
@@ -27,12 +28,12 @@ use Sub::Name;
 use Scalar::Util qw(refaddr);
 
 fieldhash my %state;
+fieldhash my %state_changed;
 fieldhash my %delayed;
-fieldhash my %delayed_ready;
 fieldhash my %delayed_once;
 
-my ( %class_isa_stateful,
-     %class_bootstrapped );
+my %class_bootstrapped;
+my %class_state_isa;
 
 sub _debug {
     my $self = shift;
@@ -42,70 +43,55 @@ sub _debug {
     warn "${class}[$addr/$state]> @_\n";
 }
 
-
 sub _state {
     my($self, $new_state) = @_;
-    $state{$self} // croak("object $self has no state, are you using Class::StateMachine::bless?");
+    $state{$self} // croak("object $self has no state, " .
+                           "use Class::StateMachine::bless to create Class::StateMachine objects");
     if (defined $new_state) {
         my $old_state = $state{$self};
-        return $old_state if $new_state eq $old_state;
+        return $old_state if $ignore_same_state_changes and $new_state eq $old_state;
         $debug and _debug($self, "changing state from $old_state to $new_state");
-        my $check = $self->can('check_state');
-        if ($check) {
+        $state_changed{$self} = 1;
+        local $state_changed{$self};
+
+        if (my $check = $self->can('check_state')) {
             $debug and _debug($self, "checking state $new_state");
             $check->($self, $new_state) or croak qq(invalid state "$new_state");
+            return $state{$self} if $state_changed{$self};
         }
-        my $leave = $self->can('leave_state');
-        if ($leave) {
+        if (my $leave = $self->can('leave_state')) {
             $debug and _debug($self, "calling leave_state($old_state, $new_state)");
             $leave->($self, $old_state, $new_state);
-            if ($state{$self} ne $old_state) {
-                $debug and _debug($self, "state transition from $old_state to $new_state shortcuted to $state{$self}");
-                return $state{$self}
-            }
+            return $state{$self} if $state_changed{$self};
         }
+
         my $base_class = ref($self);
         $base_class =~ s|::__state__::.*$||;
         my $class = _bootstrap_state_class($base_class, $new_state);
+        $state{$self} = $new_state;
         bless $self, $class;
         $debug and _debug($self, "real class set to $class");
-        $state{$self} = $new_state;
-
-        # Anything that enters the delayed array after this point
-        # should not be called until a new state change happens. We
-        # move the delayed actions to the %delayed_ready hash and call
-        # them after the enter_state method. The aim behind using the
-        # global %delayed_ready is that a state change can be induced
-        # by enter_state or from any of the delayed actions and
-        # delayed action processing restarted from there (that is
-        # here, actually!).
 
         my $delayed = $delayed{$self};
-        my $dr;
-        if ($delayed and @$delayed) {
-            delete $delayed_once{$self};
-            $dr = $delayed_ready{$self} ||= [];
-            push @$dr, @$delayed;
-            $#$delayed = -1;
-        }
+        my $delayed_top = ($delayed ? $#$delayed : -1);
 
-        my $enter = $self->can('enter_state');
-        if ($enter) {
+        if (my $enter = $self->can('enter_state')) {
             $debug and _debug($self, "calling enter_state($new_state, $old_state)");
             $enter->($self, $new_state, $old_state);
             $debug and _debug($self, "back from enter_state($new_state, $old_state)");
+            return $state{$self} if $state_changed{$self};
         }
 
-        if ($dr) {
-            while (@$dr) {
-                my $action = shift @$dr;
-                $debug and _debug($self, "running delayed action $action");
-                $self->$action
-            }
+        for (0..$delayed_top) {
+            my $action = shift @$delayed;
+            $debug and _debug($self, "running delayed action $action");
+            $self->$action;
+            return $state{$self} if $state_changed{$self};
         }
     }
+
     $debug and _debug($self, "state set to $state{$self}");
-    $state{$self};
+    return $state{$self};
 }
 
 sub _bless {
@@ -167,7 +153,6 @@ sub _bootstrap_state_class {
 
 	no strict 'refs';
         @{$state_class.'::ISA'} = $class;
-        # @{$state_class.'::ISA'} = (@on_state, $class);
         ${$state_class.'::state'} = $state;
         ${$state_class.'::base_class'} = $class;
         ${$state_class.'::state_class'} = $state_class;
@@ -218,23 +203,61 @@ sub _move_state_methods {
     }
 }
 
+sub _set_state_isa {
+    my ($class, $state, @isa) = @_;
+    %class_bootstrapped = ();
+    mro::method_changed_in($class) if mro::get_pkg_gen($class);
+    $class_state_isa{$class}{$state} = \@isa;
+}
+
+sub _state_isa {
+    my ($class, $state) = @_;
+    if (my $ref = ref $class) {
+        $state //= $class->state;
+        no strict 'refs';
+        $class = ${$ref . "::base_class"};
+    }
+    _state_isa_from_derived([grep { $_->isa('Class::StateMachine') } @{mro::get_linear_isa($class)}],
+                            $state);
+}
+
+
+sub _state_isa_from_derived {
+    my ($derived, $state) = @_;
+    my (@isa, @queue, %seen);
+    do {
+        unless ($seen{$state}++) {
+            push @isa, $state;
+            for my $class (@$derived) {
+                if (my $isa = $class_state_isa{$class}{$state}) {
+                    push @queue, @$isa;
+                    last;
+                }
+            }
+        }
+    } while (defined($state = shift @queue));
+    push @isa, '__any__';
+    wantarray ? @isa : $isa[1];
+}
+
 # use Data::Dumper;
 sub _statemachine_mro {
     my $stash = shift;
     # print Dumper $stash;
     _move_state_methods if @state_methods;
-    my $state_class = ${$stash->{state_class}};
     my $base_class = ${$stash->{base_class}};
     my $state = ${$stash->{state}};
     my @linear = @{mro::get_linear_isa($base_class)};
     my @derived = grep { $_->isa('Class::StateMachine') } @linear;
-    my @state = ($state_class,
-                 ( grep mro::get_pkg_gen($_),
-                   map(join('::', $_, '__methods__', $state   ), @derived),
-                   map(join('::', $_, '__methods__', '__any__'), @derived) ),
-                 @linear);
-    # print "mro $base_class/$state [@linear] => [@state]\n";
-    \@state;
+    my @classes;
+    for my $state (_state_isa_from_derived(\@derived, $state)) {
+        push @classes, map join('::', $_, '__methods__', $state), @derived;
+    }
+
+    # workaround bug on early mro implementations where the first
+    # class on the list returned was always discarded:
+    my $state_class = pop @classes;
+    [ $state_class, grep mro::get_pkg_gen($_), @classes, @linear ]
 }
 
 MRO::Define::register_mro('statemachine' => \&_statemachine_mro);
@@ -253,6 +276,8 @@ sub MODIFY_CODE_ATTRIBUTES {
 *bless = \&Class::StateMachine::Private::_bless;
 *delay_until_next_state = \&Class::StateMachine::Private::_delay;
 *delay_once_until_next_state = \&Class::StateMachine::Private::_delay_once;
+*set_state_isa = \&Class::StateMachine::Private::_set_state_isa;
+*state_isa = \&Class::StateMachine::Private::_state_isa;
 
 sub ref {
     my $class = ref $_[0];
@@ -339,7 +364,7 @@ Class::StateMachine - define classes for state machines
   sub new {
       my $class = shift;
       my $self = {};
-      Class::StateMachine::bless $self, $class;
+      Class::StateMachine::bless $self, $class, 'one';
       $self;
   }
 
@@ -541,8 +566,12 @@ This method calls back the methods C<check_state>, C<leave_state> and
 C<enter_state> if they are defined in the class or any of its
 subclasses for the corresponding state.
 
-If C<$new_state> equals the current object state, this method does
-nothing (including not invoking the callback methods).
+Until version 0.21, when C<$new_state> was equal to the current object
+state, this method would not invoke callback methods (C<enter_state>,
+C<leave_state>, etc.). On version 0.22 this special casing was
+removed.
+
+The variable 
 
 =over 4
 
@@ -616,6 +645,18 @@ Sets a submethod for a given class/state combination.
 
 =back
 
+=item Class::StateMachine::set_state_isa($class, $state, @isa)
+
+Allows to set one state as derived from others.
+
+Note that support for state derivation is completly experimental and
+may change at any time!
+
+=item Class::StateMachine::state_isa($class, $state)
+
+Returns the list of states from which the given C<$state> derives
+including itself and C<__any__>.
+
 =head2 Debugging
 
 Class::StateMachine supports a debugging mode that prints traces of
